from logging import getLogger
from typing import Callable, Iterable

from .config import CONFIG as _CONFIG
from .entity import Entity
from .result import SearchResultList
from .annotation import Annotations, Annotation, Correctness, Completeness


LOGGER = getLogger(__name__)


RelevanceCriterion = Callable[[Annotation], bool]


def default_relevance_criterion(annotation: Annotation) -> bool:
    if annotation.correctness == Correctness.FULL and annotation.completeness == Completeness.EXACT_OVERLAP:
        return True  # perfect match
    elif annotation.correctness == Correctness.PARTIAL and annotation.completeness == Completeness.EXACT_OVERLAP:
        return True  # different entity
    elif annotation.note == 'navíc pouze interpunknce':
        return True  # punctuation around entity
    return False


class Evaluation:
    CONFIG = _CONFIG['evaluation.Evaluation']
    TOPN = CONFIG.getint('number_of_results')

    def __init__(self, result_list: SearchResultList, annotations: Annotations,
                 entities: Iterable[Entity],
                 relevance_criterion: RelevanceCriterion = default_relevance_criterion):
        self.result_list = result_list
        self.annotations = annotations.annotations
        self.entities = list(entities)
        self.relevance_criterion = relevance_criterion

    @property
    def precision(self) -> float:
        num_relevant, num_total = 0, 0
        num_unannotated = 0
        for entity in self.entities:
            for (result_head, _) in self.result_list[entity][:self.TOPN]:
                num_total += 1
                if result_head not in self.annotations[entity]:
                    num_unannotated += 1
                    continue
                annotation = self.annotations[entity][result_head]
                if self.relevance_criterion(annotation):
                    num_relevant += 1
        if num_total == 0:
            raise ValueError('There are no results in the result list')
        if num_unannotated > 0:
            message = '{} results have not been annotated; treating them as non-relevant'.format(num_unannotated)
            LOGGER.warning(message)
        return 1.0 * num_relevant / num_total

    @property
    def recall(self) -> float:
        num_relevant, num_total = 0, 0
        num_unannotated = 0
        for entity in self.entities:
            for (result_head, _) in self.result_list[entity][:self.TOPN]:
                if result_head not in self.annotations[entity]:
                    num_unannotated += 1
                    continue
                annotation = self.annotations[entity][result_head]
                if self.relevance_criterion(annotation):
                    num_relevant += 1
            for annotation in self.annotations[entity].values():
                if self.relevance_criterion(annotation):
                    num_total += 1
        if num_unannotated > 0:
            message = '{} results have not been annotated; treating them as non-relevant'.format(num_unannotated)
            LOGGER.warning(message)
        return 1.0 * num_relevant / num_total if num_total > 0.0 else 1.0

    @property
    def f1_score(self) -> float:
        precision = self.precision
        recall = self.recall
        numerator = 2.0 * precision * recall
        denominator = precision + recall
        return numerator / denominator if denominator > 0.0 else 0.0

    def __str__(self) -> str:
        summary = 'The results achieved Precision@{topn} of {:.2f}%, Recall@{topn} of {:.2f}%, and F1@{topn}-score of {:.2f}%.'
        summary = summary.format(100.0 * self.precision, 100.0 * self.recall, 100.0 * self.f1_score, topn=self.TOPN)
        return summary

    def __repr__(self) -> str:
        return str(self)
